

#%%
from ezphot.helper import Helper

from astroquery.vizier import Vizier
from astropy.coordinates import SkyCoord
import astropy.units as u
from astroquery.gaia import Gaia
from astroquery.imcce import Skybot
#%%


class CatalogQuerier:
    """
    Class to handle external data queries using Astroquery Vizier.
    This class is a wrapper for the Astroquery Vizier and Skybot classes.
    This class providies 
    
    1. Vizier catalog query (SDSS, GAIA, 2MASS, AllWISE, PS1)
        - SDSS: SDSS DR17
        - GAIA: Gaia DR3
        - 2MASS: 2MASS All-Sky Point Source Catalog
        - AllWISE: AllWISE Data Release
        - PS1: PanSTARRS DR1
    2. Skybot catalog query (SKYBOT)
        - SKYBOT: Skybot
    
    """
    
    def __init__(self, catalog_key: str = None):
        self.skybot = Skybot
        self.vizier = Vizier()
        self.vizier.ROW_LIMIT = 1000
        self.vizier.columns = ['*', "+_r"]
        self.vizier.TIMEOUT = 60
        self.vizier.column_filters = {}
        if (catalog_key not in self.catalog_ids.keys()) and catalog_key is not None:
            raise ValueError(f"Catalog Key '{catalog_key}' is not recognized. Available keys: {list(self.catalog_ids.keys())}")
        self.current_catalog_key = catalog_key

    def __repr__(self):
        return f"CatalogQuerier(catalog={self.current_catalog_key})\n{self.config}"
    
    def query(self,
              coord,
              epoch = None,
              radius_arcsec = 3600):
        """
        Query a specific catalog around given coordinates.
        
        This method will search for the catalog in the catalog defined by the current_catalog_key.
        If the current_catalog_key is 'SKYBOT', it will use the Skybot catalog query.
        Otherwise, it will use the Vizier catalog query.
        
        Parameters
        ----------
        coord : SkyCoord
            The coordinates to query.
        epoch : Time, optional
            The epoch to query.
        radius_arcsec : float, optional
            The radius to query in arcseconds.
            
        Returns
        -------
        Table
            The catalog table.
        """
        
        
        if self.current_catalog_key is 'SKYBOT':
            return self._query_skybot_catalog(coord, epoch, radius_arcsec=radius_arcsec)
        else:
            return self._query_vizier_catalog(coord, radius_arcsec)
    
    def show_available_catalogs(self):
        """
        Display available catalogs.
        
        Returns
        -------
        list of catalog keys: list
            List of available catalogs.
        """
        print("Current catalog: ", self.current_catalog_key)
        print("Available catalogs\n==================")
        for catalog_name, catalog_id in self.catalog_ids.items():
            print(f"{catalog_name}: {catalog_id}")
        return list(self.catalog_ids.keys())
    
    def change_catalog(self, catalog_key):
        """
        Change the current catalog to query.
        
        Parameters
        ----------
        catalog_key : str
            The catalog key to change to.
        
        Returns
        -------
        None
        """
        if catalog_key in self.catalog_ids.keys():
            self.current_catalog_key = catalog_key
            print(self.__repr__())
        else:
            raise ValueError(f"Catalog Key '{catalog_key}' is not recognized.")

    def _query_vizier_catalog(self,
                             coord, 
                             radius_arcsec=10):
        """Query a specific catalog around given coordinates.
        
        Parameters
        ----------
        coord : SkyCoord
            The coordinates to query.
        radius_arcsec : float, optional
            The radius to query in arcseconds.
        """
        if type(radius_arcsec) is not u.Quantity:
            radius_arcsec = radius_arcsec * u.arcsec
        print(f'Starting query for catalog {self.current_catalog_key} around {coord} with radius {radius_arcsec}')
        print(f'{self.config}')
        result = self.vizier.query_region(coord, radius=radius_arcsec, catalog=self.current_catalog_id)
        print(f'Query completed. Found {len(result)} records.')
        return result
    
    def _query_skybot_catalog(self,
                             coord,
                             epoch = None,
                             radius_arcsec = 3600):
        """Query a specific region using Skybot.
        
        Parameters
        ----------
        coord : SkyCoord
            The coordinates to query.
        """ 
        if type(radius_arcsec) is not u.Quantity:
            radius_arcsec = radius_arcsec * u.arcsec
        if epoch is None:
            epoch = Time.now()
        print(f'Starting Skybot query around {coord} with radius {radius_arcsec}')
        result = self.skybot.cone_search(coord, radius_arcsec, epoch, location = 500)
        return result

    @property
    def config(self):
        """
        Get the configuration of the CatalogQuerier.
        
        Returns
        -------
        Configuration
            The configuration of the CatalogQuerier.
        """
        class Configuration:
            """Handles configuration for Vizier queries."""

            def __init__(self, vizier_instance: Vizier):
                self._vizier = vizier_instance

            @property
            def row_limit(self):
                return self._vizier.ROW_LIMIT

            @row_limit.setter
            def row_limit(self, value):
                self._vizier.ROW_LIMIT = value

            @property
            def columns(self):
                return self._vizier.columns

            @columns.setter
            def columns(self, value):
                self._vizier.columns = value

            @property
            def timeout(self):
                return self._vizier.TIMEOUT

            @timeout.setter
            def timeout(self, value):
                self._vizier.TIMEOUT = value

            @property
            def filters(self):
                return self._vizier.column_filters

            @filters.setter
            def filters(self, filters: dict):
                self._vizier.column_filters = filters

            def reset(self):
                self._vizier.ROW_LIMIT = -1
                self._vizier.columns = ['*']

            def __repr__(self):
                return (f"========Vizier Configuration========\n"
                        f"  row_limit       = {self.row_limit}\n"
                        f"  columns         = {self.columns}\n"
                        f"  filters         = {self.filters}\n"
                        f"  timeout         = {self.timeout} s"
                        f"\n====================================")
        
        return Configuration(self.vizier)                

    @property
    def current_catalog_id(self):
        """
        Get the current catalog ID.
        
        Returns
        -------
        str
            The current catalog ID.
        """
        if self.current_catalog_key is None:
            return None
        else:
            return self.catalog_ids[self.current_catalog_key]
    
    @property
    def catalog_ids(self):
        """
        Get the available catalog IDs.
        
        Returns
        -------
        dict
            The catalog IDs.
        """
        catalog_ids = dict()
        catalog_ids['GAIA'] = 'I/355'
        catalog_ids['GAIA_DR3'] = "I/355/gaiadr3"
        catalog_ids['GAIA_DR3_SPEC'] = 'I/355/spectra'
        catalog_ids['GAIAXP'] = 'I/355/xpsample'      
        # 2MASS (Final release)
        catalog_ids['2MASS'] = "II/246/out"

        # AllWISE (All-sky WISE data)
        catalog_ids['AllWISE'] = "II/328/allwise"

        # Pan-STARRS DR1 (Stacked photometry)
        catalog_ids['PS1'] = "II/349/ps1"

        # SDSS DR17 (Photometric data)
        catalog_ids['SDSS'] = "V/167/sdss17"
        
        # Skybot
        catalog_ids['SKYBOT'] = "IMCCE/Skybot"
        return catalog_ids
        
#%%
if __name__ == "__main__":
    from astropy.time import Time
    # Example usage
    self = CatalogQuerier(catalog_key='SKYBOT')
    
    # Query a specific region
    coord = SkyCoord(ra=233.322342*u.deg, dec=-68.007909*u.deg, frame='icrs')
    epoch = Time('2025-02-09T00:00:00')
    #result = ed.query_catalog(coord, radius_arcsec=5)
    
    result = self.query(coord = coord, epoch = epoch)
# %%





#%%
from pathlib import Path
from astropy.time import Time
from astropy.io import ascii
from astropy.table import Table, vstack
from typing import List, Dict, Union
import ccdproc
from ccdproc import CCDData
import astropy.units as u
import numpy as np
import glob, os
from tqdm import tqdm
from functools import reduce
from datetime import datetime
import gc
import psutil, os
import tracemalloc

from ezphot.helper import Helper
from ezphot.imageobjects import ScienceImage
from ezphot.imageobjects import CalibrationImage
from ezphot.methods import Stack

#%%
class Preprocess:
    """
    Preprocess the image.
    
    This class provides methods 
    
    1. Get master frame from the image.
    
    2. Correct bias, dark, and flat.
    
    3. Correct bias and dark.
    
    4. Correct flat.
    
    """
    
    def __init__(self):
        self.helper = Helper()
        self._cached_masterframe_tbl = None
        
    def get_masterframe_from_image(self,
                                   target_img: ScienceImage or CalibrationImage,
                                   imagetyp: str = 'BIAS',
                                   max_days: float = 5):
        """
        Get master frame from the image.
        
        This method will search for the master frame in the master frame directory.
        
        Parameters
        ----------
        target_img : ScienceImage or CalibrationImage
            The target image to get the master frame from.
        imagetyp : str, optional
            The type of image to get the master frame from.
        max_days : float, optional
            The maximum number of days to search for the master frame.
            
        Returns
        -------
        master_frames : Table
            Metadata of the master frame(s) found.
        """
        if not target_img.is_header_loaded:
            target_img.header
        master_frames = self.get_masterframe(
            observatory= target_img.observatory,
            telkey = target_img.telkey,
            telname = target_img.telname,
            imagetyp = imagetyp,
            exptime = target_img.exptime,
            filter_ = target_img.filter,
            obsdate = target_img.obsdate,
            max_days = max_days)
        return master_frames
        
    def get_masterframe(self,
                        observatory: str = '7DT',
                        telkey: str = '7DT_C361K_HIGH_1x1',
                        telname: str = '7DT02',
                        imagetyp: str = 'BIAS',
                        exptime: float = None,
                        filter_: str = None,
                        obsdate: str = None,
                        max_days: float = 5):
        """
        Get master frame from the image.
        
        This method will search for the master frame in the master frame directory.
        
        Parameters
        ----------
        observatory : str, optional
            The observatory to get the master frame from.
        telkey : str, optional
            The telescope key to get the master frame from.
        telname : str, optional
            The telescope name to get the master frame from.
        imagetyp : str, optional
            The type of image to get the master frame from.
        exptime : float, optional
            The exposure time of the image to get the master frame from.
        filter_ : str, optional
            The filter of the image to get the master frame from.
        obsdate : str, optional
            The observation date of the image to get the master frame from.
        max_days : float, optional
            The maximum number of days to search for the master frame.
            
        Returns
        -------
        master_frames : Table
            Metadata of the master frame(s) found.
        """
        from astropy.time import Time
        import numpy as np

        # Load summary tables
        all_masterframe_info = {}
        masterframe_summary_path = Path(self.config['CALIBDATA_MASTERDIR']) / 'summary.ascii_fixed_width'
        
        if self._cached_masterframe_tbl is None:
            if masterframe_summary_path.exists():
                all_masterframe_tbl = ascii.read(masterframe_summary_path, format='fixed_width')
            else:
                print(f"[WARNING] Master frame summary file not found: {masterframe_summary_path}")
                return 
            self._cached_masterframe_tbl = all_masterframe_tbl

        all_masterframe_tbl = self._cached_masterframe_tbl

        if len(all_masterframe_tbl) == 0:
            raise FileNotFoundError("No calibration frame metadata found.")
        
        if obsdate is None:
            print('No obsdate provided. Returning None')
            return

        # === Base mask ===
        mask = np.ones(len(all_masterframe_tbl), dtype=bool)

        # Apply filters
        if observatory is not None:
            mask &= (all_masterframe_tbl['observatory'] == observatory)
        if telkey is not None:
            mask &= (all_masterframe_tbl['telkey'] == telkey)
        if telname is not None:
            mask &= (all_masterframe_tbl['telname'] == telname)
        if imagetyp is not None:
            mask &= (all_masterframe_tbl['imagetyp'] == imagetyp.upper())

        if imagetyp.upper() == 'DARK' and exptime is not None and 'exptime' in all_masterframe_tbl.colnames:
            mask &= np.isclose(all_masterframe_tbl['exptime'], exptime)
        elif imagetyp.upper() == 'FLAT' and filter_ is not None and 'filtername' in all_masterframe_tbl.colnames:
            mask &= (all_masterframe_tbl['filtername'] == filter_)

        all_masterframe_tbl = all_masterframe_tbl[mask]
        # === group_id logic ===
        use_group_id = False
        valid_group_id = None

        obsdate_target = self.helper.flexible_time_parser(obsdate).mjd
        obs_times = Time(all_masterframe_tbl['obsdate'], format='isot').mjd
        delta_t = np.abs(obs_times - obsdate_target)
        all_masterframe_tbl['time_difference'] = delta_t

        if 'group_id' in all_masterframe_tbl.colnames:
            # Try to find the nearest row with a valid group_id
            sorted_indices = np.argsort(delta_t)
            for idx in sorted_indices:
                gid = all_masterframe_tbl['group_id'][idx]
                if gid >= 0:
                    valid_group_id = gid
                    use_group_id = True
                    break  # stop at the first valid group_id

        if use_group_id:
            sub_mask = (all_masterframe_tbl['group_id'] == valid_group_id)
        else:
            # Fallback to time-based filtering
            delta_t = all_masterframe_tbl['time_difference']
            within_range = delta_t < max_days
            if not np.any(within_range):
                print("No calibration frame found within max_days.")
                return None
            sub_mask = within_range

        # === Final filtering ===
        filtered_tbl = all_masterframe_tbl[sub_mask]

        if len(filtered_tbl) == 0:
            print("No calibration frame found with the specified criteria.")
            return None

        filtered_tbl.sort('time_difference')

        return filtered_tbl

    def correct_bdf(self, target_img: ScienceImage, 
                    bias_image: CalibrationImage, 
                    dark_image: CalibrationImage, flat_image: CalibrationImage,
                    save : bool = False,
                    **kwargs
                    ):
        """
        Correct bias, dark, and flat.
        
        Parameters
        ----------
        target_img : ScienceImage
            The target image to correct bias, dark, and flat.
        bias_image : CalibrationImage
            The bias image to correct bias, dark, and flat.
        dark_image : CalibrationImage
            The dark image to correct bias, dark, and flat.
        flat_image : CalibrationImage
            The flat image to correct bias, dark, and flat.
        save : bool, optional
            Whether to save the corrected image.
        **kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """
        # if target_img.status.BIASCOR['status']:
        #     target_img.logger.warning(f"BIAS correction already applied to {target_img.path}. BIAS correction is not applied.")
        #     raise RuntimeError(f"BIAS correction already applied to {target_img.path}")
        # if target_img.status.DARKCOR['status']:
        #     target_img.logger.warning(f"DARK correction already applied to {target_img.path}. BIAS correction is not applied.")
        #     raise RuntimeError(f"DARK correction already applied to {target_img.path}")
        # if target_img.status.FLATCOR['status']:
        #     target_img.logger.warning(f"FLAT correction already applied to {target_img.path}. BIAS correction is not applied.")
        #     raise RuntimeError(f"FLAT correction already applied to {target_img.path}")

        # Convert input images to CCDData
        if target_img.data is None:
            target_img.load_data()
        if bias_image.data is None:
            bias_image.load_data()
        if dark_image.data is None:
            dark_image.load_data()
        if flat_image.data is None:
            flat_image.load_data()
        sci_ccddata = ccdproc.CCDData(data = target_img.data, meta = target_img.header, unit = 'adu')
        bias_ccddata = ccdproc.CCDData(data = bias_image.data, meta = bias_image.header, unit = 'adu')
        dark_ccddata = ccdproc.CCDData(data = dark_image.data, meta = dark_image.header, unit = 'adu')
        flat_ccddata = ccdproc.CCDData(data = flat_image.data, meta = flat_image.header, unit = 'adu')

        # Perform bias, dark, flat correction
        calib_data = self._correct_bdf(tgt_data = sci_ccddata, bias_data = bias_ccddata, dark_data = dark_ccddata, flat_data = flat_ccddata)        

        # Determine data types and convert to selected data type
        tgt_dtype = target_img.data.dtype
        bias_dtype = bias_image.data.dtype
        dark_dtype = dark_image.data.dtype
        flat_dtype = flat_image.data.dtype
        selected_dtype = reduce(np.promote_types, [tgt_dtype, bias_dtype, dark_dtype, flat_dtype])
        calib_data.data = calib_data.data.astype(selected_dtype)
        
        # Add metadata
        update_header_kwargs = dict(
            BIASCOR = True,
            BCORTIME = Time.now().isot,
            BIASPATH = str(bias_image.savepath.savepath),
            DARKCOR = True, 
            DCORTIME = Time.now().isot,
            DARKPATH = str(dark_image.savepath.savepath),
            FLATCOR = True,
            FCORTIME = Time.now().isot,
            FLATPATH = str(flat_image.savepath.savepath),
        )
        
        
        calib_data.header.update(update_header_kwargs)

        # Create new image object
        calib_image = type(target_img)(path  = target_img.path, telinfo = target_img.telinfo, status = target_img.status, load = False)
        
        calib_image.data = calib_data.data
        calib_image.header = calib_data.header
        calib_image.update_status(process_name= 'BIASCOR')
        calib_image.update_status(process_name = 'DARKCOR')
        calib_image.update_status(process_name = 'FLATCOR')
        # calib_image.logger.info(f"BIAS correction applied with {bias_image.savepath.savepath}")
        # calib_image.logger.info(f"DARK correction applied with {dark_image.savepath.savepath}")
        # calib_image.logger.info(f"FLAT correction applied with {flat_image.savepath.savepath}")
        # bias_image.logger.info(f"Used for BIAS correction: FILEPATH = {calib_image.savepath.savepath}")
        # dark_image.logger.info(f"Used for DARK correction: FILEPATH = {calib_image.savepath.savepath}")
        # flat_image.logger.info(f"Used for FLAT correction: FILEPATH = {calib_image.savepath.savepath}")
        if save:
            calib_image.write()
        return calib_image
    
    def _correct_bdf(self, tgt_data : CCDData, bias_data : CCDData, dark_data : CCDData, flat_data : CCDData):
        bcalib_data = self._correct_bias(tgt_data = tgt_data, bias_data = bias_data)
        dbcalib_data = self._correct_dark(tgt_data = bcalib_data, dark_data = dark_data)
        fdbcalib_data = self._correct_flat(tgt_data = dbcalib_data, flat_data = flat_data)
        return fdbcalib_data
    
    def correct_bd(self, target_img: ScienceImage, 
                   bias_image: CalibrationImage, 
                   dark_image: CalibrationImage, 
                   save : bool = False,
                   **kwargs
                   ):
        """
        Correct bias and dark.
        
        Parameters
        ----------
        target_img : ScienceImage   
            The target image to correct bias and dark.
        bias_image : CalibrationImage
            The bias image to correct bias and dark.
        dark_image : CalibrationImage
            The dark image to correct bias and dark.
        save : bool, optional
            Whether to save the corrected image.
        **kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """

        # Convert input images to CCDData
        if target_img.data is None:
            target_img.load_data()
        if bias_image.data is None:
            bias_image.load_data()
        if dark_image.data is None:
            dark_image.load_data()
        sci_ccddata = ccdproc.CCDData(data = target_img.data, meta = target_img.header, unit = 'adu')
        bias_ccddata = ccdproc.CCDData(data = bias_image.data, meta = bias_image.header, unit = 'adu')
        dark_ccddata = ccdproc.CCDData(data = dark_image.data, meta = dark_image.header, unit = 'adu')

        # Perform bias, dark correction
        calib_data = self._correct_bd(tgt_data = sci_ccddata, bias_data = bias_ccddata, dark_data = dark_ccddata)

        # Determine data types and convert to selected data type
        tgt_dtype = target_img.data.dtype
        bias_dtype = bias_image.data.dtype
        dark_dtype = dark_image.data.dtype
        selected_dtype = reduce(np.promote_types, [tgt_dtype, bias_dtype, dark_dtype])
        #selected_dtype = np.promote_types(tgt_dtype, bias_dtype, dark_dtype)
        calib_data.data = calib_data.data.astype(selected_dtype)
        
        # Add metadata
        update_header_kwargs = dict(
            BIASCOR = True,
            BCORTIME = Time.now().isot,
            BIASPATH = str(bias_image.savepath.savepath),
            DARKCOR = True, 
            DCORTIME = Time.now().isot,
            DARKPATH = str(dark_image.savepath.savepath),
        )
        
        calib_data.header.update(update_header_kwargs)

        # Create new image object
        calib_image = type(target_img)(path  = target_img.path, telinfo = target_img.telinfo, status = target_img.status, load = False)
        calib_image.data = calib_data.data
        calib_image.header = calib_data.header
        calib_image.update_status(process_name= 'BIASCOR')
        calib_image.update_status(process_name = 'DARKCOR')
        # calib_image.logger.info(f"BIAS correction applied with {bias_image.savepath.savepath}")
        # calib_image.logger.info(f"DARK correction applied with {dark_image.savepath.savepath}")
        # bias_image.logger.info(f"Used for BIAS correction: FILEPATH = {calib_image.savepath.savepath}")
        # dark_image.logger.info(f"Used for DARK correction: FILEPATH = {calib_image.savepath.savepath}")
        if save:
            calib_image.write()
        
        return calib_image
    
    def _correct_bd(self, tgt_data : CCDData, bias_data : CCDData, dark_data : CCDData):
        bcalib_data = self._correct_bias(tgt_data = tgt_data, bias_data = bias_data)
        dbcalib_data = self._correct_dark(tgt_data = bcalib_data, dark_data = dark_data)
        return dbcalib_data
        
    def correct_bias(self, target_img: ScienceImage or CalibrationImage, 
                     bias_image: CalibrationImage,
                     save : bool = False,
                     **kwargs
                     ):
        """
        Correct bias in the image.
        
        Parameters
        ----------
        target_img : ScienceImage or CalibrationImage
            The target image to correct bias.
        bias_image : CalibrationImage
            The bias image to correct bias.
        save : bool, optional
            Whether to save the corrected image.
        **kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """
        # if target_img.status.BIASCOR['status']:
        #     target_img.logger.warning(f"BIAS correction already applied to {target_img.path}. BIAS correction is not applied.")
        #     raise RuntimeError(f"BIAS correction already applied to {target_img.path}")
        
        # Convert input images to CCDData
        if target_img.data is None:
            target_img.load_data()
        if bias_image.data is None:
            bias_image.load_data()
        sci_ccddata = ccdproc.CCDData(data = target_img.data, meta = target_img.header, unit = 'adu')
        bias_ccddata = ccdproc.CCDData(data = bias_image.data, meta = bias_image.header, unit = 'adu')
        
        # Perform bias correction
        calib_data = self._correct_bias(tgt_data = sci_ccddata, bias_data = bias_ccddata)
        
        # Determine data types and convert to selected data type
        tgt_dtype = target_img.data.dtype
        bias_dtype = bias_image.data.dtype
        selected_dtype = np.promote_types(tgt_dtype, bias_dtype)
        calib_data.data = calib_data.data.astype(selected_dtype)

        # Update header
        update_header_kwargs = dict(
            BIASCOR = True,
            BCORTIME = Time.now().isot,
            BIASPATH = str(bias_image.savepath.savepath),
        )
        
        calib_data.header.update(update_header_kwargs)
        
        # Create new image object
        calib_image = type(target_img)(path  = target_img.path, telinfo = target_img.telinfo, status = target_img.status, load = False)
        calib_image.data = calib_data.data
        calib_image.header = calib_data.header
        calib_image.update_status(process_name= 'BIASCOR')
        # calib_image.logger.info(f"BIAS correction applied with {bias_image.savepath.savepath}")
        # bias_image.logger.info(f"Used for BIAS correction: FILEPATH = {calib_image.savepath.savepath}")
        if save:
            calib_image.write()
        return calib_image

    
    def _correct_bias(self, tgt_data : CCDData, bias_data : CCDData):
        calib_data = ccdproc.subtract_bias(tgt_data, bias_data)
        return calib_data
    
    def correct_dark(self, target_img: ScienceImage or CalibrationImage, 
                     dark_image: CalibrationImage, 
                     save : bool = False,
                     **kwargs
                     ):
        """
        Correct dark in the image.
        
        Parameters
        ----------
        target_img : ScienceImage or CalibrationImage
            The target image to correct dark.
        dark_image : CalibrationImage
            The dark image to correct dark.
        save : bool, optional
            Whether to save the corrected image.
        **kwargs : dict, optional
            Additional keyword arguments.
        
        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """

        # Convert input images to CCDData
        if target_img.data is None:
            target_img.load_data()
        if dark_image.data is None:
            dark_image.load_data()
        sci_ccddata = ccdproc.CCDData(data = target_img.data, meta = target_img.header, unit = 'adu')
        dark_ccddata = ccdproc.CCDData(data = dark_image.data, meta = dark_image.header, unit = 'adu')
        
        # Perform dark correction
        calib_data = self._correct_dark(tgt_data = sci_ccddata, dark_data = dark_ccddata)
        
        # Determine data types and convert to selected data type
        tgt_dtype = target_img.data.dtype
        dark_dtype = dark_image.data.dtype
        selected_dtype = np.promote_types(tgt_dtype, dark_dtype)
        calib_data.data = calib_data.data.astype(selected_dtype)
        
        # Add metadata
        update_header_kwargs = dict(
            DARKCOR = True,
            DCORTIME = Time.now().isot,
            DARKPATH = str(dark_image.savepath.savepath),
        )
        
        calib_data.header.update(update_header_kwargs)
        
        # Create new image object
        calib_image = type(target_img)(path  = target_img.path, telinfo = target_img.telinfo, status = target_img.status, load = False)
        calib_image.data = calib_data.data
        calib_image.header = calib_data.header
        calib_image.update_status(process_name= 'DARKCOR')
        # calib_image.logger.info(f"DARK correction applied with {dark_image.savepath.savepath}")
        # dark_image.logger.info(f"Used for DARK correction: FILEPATH = {calib_image.savepath.savepath}")
        if save:
            calib_image.write()
        return calib_image

    def _correct_dark(self, tgt_data : CCDData, dark_data : CCDData):
        calib_data = ccdproc.subtract_dark(tgt_data, dark_data, scale = True, exposure_time = 'EXPTIME', exposure_unit = u.second)
        return calib_data
    
    def correct_flat(self, target_img: ScienceImage, flat_image: CalibrationImage,
                     save : bool = False,
                     **kwargs
                     ):
        """
        Correct flat in the image.
        
        Parameters
        ----------
        target_img : ScienceImage
            The target image to correct flat.
        flat_image : CalibrationImage
            The flat image to correct flat.
        save : bool, optional
            Whether to save the corrected image.
        **kwargs : dict, optional
            Additional keyword arguments.

        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """
        # Convert input images to CCDData
        if target_img.data is None:
            target_img.load_data()
        if flat_image.data is None:
            flat_image.load_data()
        sci_ccddata = ccdproc.CCDData(data = target_img.data, meta = target_img.header, unit = 'adu')
        flat_ccddata = ccdproc.CCDData(data = flat_image.data, meta = flat_image.header, unit = 'adu')
        
        # Perform flat correction
        calib_data = self._correct_flat(tgt_data = sci_ccddata, flat_data = flat_ccddata)
        
        # Determine data types and convert to selected data type
        tgt_dtype = target_img.data.dtype
        flat_dtype = flat_image.data.dtype
        selected_dtype = np.promote_types(tgt_dtype, flat_dtype)
        calib_data.data = calib_data.data.astype(selected_dtype)
        
        # Update header
        update_header_kwargs = dict(
            FLATCOR = True,
            FCORTIME = Time.now().isot,
            FLATPATH = str(flat_image.savepath.savepath),
        )
        
        calib_data.header.update(update_header_kwargs)
        
        # Create new image object
        calib_image = type(target_img)(path  = target_img.path, telinfo = target_img.telinfo, status = target_img.status, load = False)
        calib_image.data = calib_data.data
        calib_image.header = calib_data.header
        calib_image.update_status(process_name= 'FLATCOR')
        # calib_image.logger.info(f"FLAT correction applied with {flat_image.savepath.savepath}")
        # flat_image.logger.info(f"Used for FLAT correction: FILEPATH = {calib_image.savepath.savepath}")
        if save:
            calib_image.write()
        return calib_image
        
    def _correct_flat(self, tgt_data : CCDData, flat_data : CCDData):
        calib_data = ccdproc.flat_correct(tgt_data, flat_data)
        return calib_data
    
    def generate_masterframe(self, calib_imagelist : List[CalibrationImage], 
                             mbias : Union[CalibrationImage or List[CalibrationImage]] = None,
                             mdark : Union[CalibrationImage or List[CalibrationImage]] = None,
                             
                             # Combine parameters
                             combine_type: str = 'median',
                             n_proc: int = 4,
                             clip_type: str = 'extrema',
                             sigma: float = 3.0,
                             nlow: int = 1,
                             nhigh: int = 1,
                            
                             # Other parameters
                             verbose: bool = True,
                             save: bool = True,
                             **kwargs
                             ):
        """
        Generate master bias, dark, flat frames.
        
        Parameters
        ----------
        calib_imagelist : List[CalibrationImage]
            The list of calibration images to generate master frames from.
        mbias : Union[CalibrationImage or List[CalibrationImage]], optional
            The master bias image to use for bias correction.
        mdark : Union[CalibrationImage or List[CalibrationImage]], optional
            The master dark image to use for dark correction.
        combine_type : str, optional
            The type of combination to use for the master frames.
        n_proc : int, optional
            The number of processors to use for the master frames.
        clip_type : str, optional
            The type of clipping to use for the master frames.
        sigma : float, optional
            The sigma for the master frames.
        nlow : int, optional
            The number of low values to clip for the master frames.
        nhigh : int, optional
            The number of high values to clip for the master frames.
        verbose : bool, optional
            Whether to print verbose output.
        save : bool, optional
            Whether to save the master frames.
        **kwargs : dict, optional   
            Additional keyword arguments.

        Returns
        -------
        master_files : dict
            The master frames.
        """
        
        def empty_memory(imagelist : Union[CalibrationImage or List[CalibrationImage]]):
            if isinstance(imagelist, CalibrationImage):
                imagelist = [imagelist]
            for image in imagelist:
                image.data = None

        combiner = Stack()
        all_filelist = [image.path for image in calib_imagelist]
        all_fileinfo = self.helper.get_imginfo(all_filelist, normalize_key = True)
        all_fileinfo['image'] = calib_imagelist
        all_fileinfo_by_group = all_fileinfo.group_by(['binning', 'gain']).groups
        master_files = dict()
        for group in all_fileinfo_by_group:
            key = (group['binning'][0], group['gain'][0])
            master_files[key] = dict(BIAS = None, DARK = dict(), FLAT = dict())
        
        if mbias:
            if isinstance(mbias, CalibrationImage):
                mbias = [mbias]
            for bias in mbias:
                bias_key = (str(bias.binning), str(bias.gain))
                master_files[bias_key]['BIAS'] = bias
        if mdark:
            if isinstance(mdark, CalibrationImage):
                mdark = [mdark]
            for dark in mdark:
                header = dark.header
                dark_key = (str(dark.binning), str(dark.gain))
                master_files[dark_key]['DARK'][str(dark.exptime)] = dark
        
        # Run the calibration
        for group in all_fileinfo_by_group:
            # Separate the images by type
            key = (group['binning'][0], group['gain'][0])
            if not master_files[key]['BIAS']:
                bias_key = ['BIAS', 'ZERO']
                bias_mask  = np.isin(group['imgtype'], bias_key)
                bias_fileinfo = group[bias_mask]
                new_bias = None
                if bias_fileinfo:
                    bias_rep = bias_fileinfo[0]
                    bias_imagelist = bias_fileinfo['image']
                    new_bias, _ = combiner.stack_multiprocess(
                        target_imglist = bias_imagelist,
                        target_errormaplist= None,
                        target_outpath = None,
                        weight_outpath = None,
                        combine_type = combine_type,
                        n_proc = n_proc,
                        clip_type = clip_type,
                        sigma = sigma, 
                        nlow = nlow,
                        nhigh = nhigh,
                        verbose = verbose)
                    empty_memory(bias_imagelist)
                    
                    master_files[key]['BIAS'] = new_bias
                else:
                    new_bias = self.get_masterframe_from_image(target_img = group[0]['image'], imagetyp = 'BIAS')
                    master_files[key]['BIAS'] = new_bias
                                
                if save:
                    new_bias.write()
                
            if not master_files[key]['DARK']:
                if not master_files[key]['BIAS']:
                    raise ValueError("Master BIAS is not found.")
                
                dark_key = ['DARK']
                dark_mask  = np.isin(group['imgtype'], dark_key)
                dark_fileinfo = group[dark_mask]
                if dark_fileinfo:
                    dark_fileinfo_by_exptime = dark_fileinfo.group_by('exptime').groups
                    for dark_group in dark_fileinfo_by_exptime:
                        dark_rep = dark_group[0]
                        exptime_name = dark_rep['exptime']
                        b_darkimagelist = []
                        dark_imagelist = dark_group['image']
                        for dark in tqdm(dark_imagelist, desc = 'BIAS correction on DARK frames...'):
                            b_dark_image = self.correct_bias(
                                target_img = dark, 
                                bias_image = master_files[key]['BIAS'],
                                save = False)
                            b_darkimagelist.append(b_dark_image)
                        
                        empty_memory(dark_imagelist)
                        
                        del dark_imagelist
                        gc.collect()
                        
                        new_dark, _ = combiner.stack_multiprocess(
                            target_imglist = b_darkimagelist,
                            target_errormaplist= None,
                            target_outpath = None,
                            weight_outpath = None,
                            combine_type = combine_type,
                            n_proc = n_proc,
                            clip_type = clip_type,
                            sigma = sigma, 
                            nlow = nlow,
                            nhigh = nhigh,
                            verbose = verbose)   
                        master_files[key]['DARK'][exptime_name] = new_dark

                        if save:
                            new_dark.write()
                        
                        del b_darkimagelist
                        gc.collect()       
                  

                if '100.0' not in master_files[key]['DARK'].keys():
                    new_dark = self.get_masterframe(target_img = group[0]['image'],
                                                    imagetyp = 'DARK',
                                                    exptime = 100)
                    master_files[key]['DARK']['100.0'] = new_dark
                
            
            if not master_files[key]['FLAT']:    
                if (not master_files[key]['DARK']['100.0']) or (not master_files[key]['BIAS']):
                    raise ValueError("Master BIAS or DARK frame not found.")

                flat_key = ['FLAT']
                flat_mask = np.isin(group['imgtype'], flat_key)
                flat_fileinfo = group[flat_mask]
                if flat_fileinfo:
                    flat_fileinfo_by_filter = flat_fileinfo.group_by('filter').groups
                    for flat_group in flat_fileinfo_by_filter:
                        flat_rep = flat_group[0]
                        filter_name = flat_rep['filter']
                        db_flatimagelist = []
                        flat_imagelist = flat_group['image']
                        for flat in tqdm(flat_imagelist, desc = 'BIAS, DARK correction on FLAT frames...'):
                            mbias = master_files[key]['BIAS']
                            # Convert list of exposure time strings to floats
                            mdark_exptime_key_float = [float(x) for x in master_files[key]['DARK'].keys()]

                            # Find the closest exposure time to 20.0 seconds
                            closest_exptime = min(mdark_exptime_key_float, key=lambda x: abs(x - 20.0))

                            # Use the string version as key to access master dark
                            closest_exptime_key = str(closest_exptime)       
                            mdark = master_files[key]['DARK'][closest_exptime_key]  
                                               
                            db_flat_image = self.correct_bd(
                                target_img = flat, 
                                bias_image = mbias, 
                                dark_image = mdark,
                                save = False)
                            db_flatimagelist.append(db_flat_image)
                        empty_memory(flat_imagelist)
                        
                        new_flat, _ = combiner.stack_multiprocess(
                            target_imglist = db_flatimagelist,
                            target_errormaplist= None,
                            target_outpath = None,
                            weight_outpath = None,
                            combine_type = combine_type,
                            n_proc = n_proc,
                            clip_type = clip_type,
                            sigma = sigma, 
                            nlow = nlow,
                            nhigh = nhigh,
                            verbose = verbose)    
                        master_files[key]['FLAT'][filter_name] = new_flat
                        if save:
                            new_flat.write()
                        
                        del db_flatimagelist
                        gc.collect()    
            
        return master_files

    # def generate_bpmask(self, data, output, n_sigma=5, header=None):
    #     """
    #     Expects 2D cupy array as data (master dark)
    #     Func to generate hot pixel mask.
    #     Bad pixel criterion set by J.H. Bae
    #     """
    #     # from astropy.stats import sigma_clipped_stats
    #     # from eclaire.stats import sigma_clipped_stats

    #     # if self.queue:
    #     #     self.queue.log_memory_stats(f"After loading {dtype} data")

    #     # mean, median, std = sigma_clipped_stats(data, sigma=3, maxiters=5) # astropy
    #     # median, std = sigma_clipped_stats(data, reduce="median", width=3, iters=5) # eclaire
    #     mean, median, std = sigma_clipped_stats_cupy(data, sigma=3, maxiters=5)

    #     hot_mask = cp.abs(data - median) > n_sigma * std  # 1 for bad, 0 for okay
    #     hot_mask = cp.asnumpy(hot_mask).astype("uint8")

    #     # # Save the file
    #     # newhdu = fits.PrimaryHDU(hot_mask)

    #     # Use CompImageHDU for less disk consumption
    #     newhdu = fits.CompImageHDU(data=hot_mask)

    #     # Create a primary HDU as placeholder
    #     primary_hdu = fits.PrimaryHDU()

    #     if header:
    #         # for card in header.cards:
    #         #     # skip certain keywords
    #         #     if card.keyword in ("BITPIX", "NAXIS", "NAXIS1", "NAXIS2"):
    #         #         continue
    #         #     newhdu.header.append(card)
    #         for key in [
    #             "INSTRUME",
    #             "GAIN",
    #             "EXPTIME",
    #             "EXPOSURE",
    #             "JD",
    #             "MJD",
    #             "DATE-OBS",
    #             "DATE-LOC",
    #             "XBINNING",
    #             "YBINNING",
    #         ]:
    #             newhdu.header[key] = header[key]
    #         # newhdu.header.add_comment("Above header is from the first dark frame")
    #         newhdu.header["COMMENT"] = "Header inherited from first dark frame"
    #     newhdu.header["NHOTPIX"] = (np.sum(hot_mask), "Number of hot pixels.")
    #     # newhdu.header['NDARKIMG'] = (len(bias_sub_dark_names), 'Number of bias subtracted dark images used in sigma-clipped mean combine.')
    #     # newhdu.header['EXPTIME'] = dark_hdr['EXPTIME']
    #     # newhdu.header['COMMENT'] = 'Stable hot pixel mask image. Algorithm produced by J.H.Bae on 20250117.'
    #     newhdu.header["SIGMAC"] = (n_sigma, "HP threshold in clipped sigma")
    #     newhdu.header["BADPIX"] = (1, "Pixel Value for Bad pixels")

    #     # newhdul = fits.HDUList([newhdu])
    #     newhdul = fits.HDUList([primary_hdu, newhdu])
    #     newhdul.writeto(output, overwrite=True)

    #     # if self.queue:
    #     #     self.queue.log_memory_stats(f"After saving {dtype}")

    # def interpolate_masked_pixels_gpu_vectorized_weight(
    #     image,
    #     mask,
    #     weight=None,
    #     window=1,
    #     method="median",
    #     badpix=1,
    # ):
    #     """Interpolate masked pixels in image (and optional weight map) using a mean filter.

    #     Args:
    #         image (cp.ndarray): 2D input image.
    #         mask (cp.ndarray): Binary mask (1 = masked, 0 = valid) when badpix=1.
    #         weight (cp.ndarray, optional): Pixel weight map. If None, unweighted mean is used. It must be 1/VARIANCE.
    #         window (int): Radius of square window.
    #         method (str): Interpolation method. Options are "inverse_variance" or "median".

    #     Returns:
    #         interpolated_image (cp.ndarray): Image with interpolated masked pixels.
    #         interpolated_weight (cp.ndarray): Weight map with interpolated weights.
    #     """
    #     import cupy as cp

    #     H, W = image.shape
    #     assert image.shape == mask.shape
    #     if weight is not None:
    #         assert weight.shape == image.shape

    #     result = image.copy()
    #     weight_result = weight.copy() if weight is not None else None  # cp.ones_like(image)

    #     ys, xs = cp.where(mask == badpix)
    #     N = ys.shape[0]
    #     if N == 0:
    #         return result, weight_result

    #     # Create patch offsets
    #     dy, dx = cp.meshgrid(
    #         cp.arange(-window, window + 1), cp.arange(-window, window + 1), indexing="ij"
    #     )
    #     dy = dy.ravel()
    #     dx = dx.ravel()

    #     patch_ys = ys[:, None] + dy[None, :]
    #     patch_xs = xs[:, None] + dx[None, :]

    #     # Mask out-of-bound locations
    #     in_bounds = (patch_ys >= 0) & (patch_ys < H) & (patch_xs >= 0) & (patch_xs < W)

    #     # Clip for safe indexing
    #     patch_ys_safe = cp.clip(patch_ys, 0, H - 1)
    #     patch_xs_safe = cp.clip(patch_xs, 0, W - 1)
    #     flat_indices = patch_ys_safe * W + patch_xs_safe

    #     flat_image = image.ravel()
    #     flat_mask = mask.ravel()
    #     patch_vals = flat_image[flat_indices]
    #     patch_mask = flat_mask[flat_indices]

    #     valid = (patch_mask == 1 - badpix) & in_bounds

    #     if weight is not None:
    #         flat_weight = weight.ravel()
    #         patch_weights = flat_weight[flat_indices]

    #         patch_weights = cp.where(valid, patch_weights, 0)
    #         patch_vals = cp.where(valid, patch_vals, cp.nan)

    #         if method == "inverse_variance":
    #             weighted_sum = cp.sum(patch_weights * patch_vals, axis=1)
    #             weight_total = cp.sum(patch_weights, axis=1)

    #             # if weight_total == 0, val is 0
    #             interp_vals = cp.where(weight_total > 0, weighted_sum / weight_total, 0)
    #             interp_weights = weight_total  # assuming weight = 1/var

    #         elif method == "median":
    #             # Compute median and MAD
    #             interp_vals = cp.nanmedian(patch_vals, axis=1)

    #             # Handle NaNs safely by using cp.isclose (for float stability)
    #             mask = cp.isclose(patch_vals, interp_vals[:, None], equal_nan=True)

    #             # For multiple matches per row, take first match
    #             def first_match_per_row(data, mask):
    #                 idx = cp.argmax(mask, axis=1)  # index of first True in each row
    #                 row_idx = cp.arange(mask.shape[0])
    #                 return data[row_idx, idx]

    #             interp_weights = first_match_per_row(patch_weights, mask)

    #         # Fill interpolated values and weights
    #         result[ys, xs] = interp_vals
    #         weight_result[ys, xs] = interp_weights

    #         return result, weight_result

    #     else:
    #         patch_vals = cp.where(valid, patch_vals, cp.nan)
    #         interp_vals = cp.nanmedian(patch_vals, axis=1)

    #         result[ys, xs] = interp_vals
    #         return result  


# %%




#%%
from pathlib import Path
from tqdm import tqdm
import pandas as pd
from typing import Optional, Union, List
from astropy.io import fits
from astropy.time import Time
from astropy.table import Table, vstack, join
from astropy.coordinates import SkyCoord
import astropy.units as u
from concurrent.futures import ProcessPoolExecutor
import numpy as np

from ezphot.configuration import TIPConfig
from ezphot.dataobjects import Catalog
from ezphot.imageobjects import ScienceImage
from ezphot.helper import Helper

#%%

class CatalogSet:
    """
    CatalogSet class for managing a set of catalogs.
    
    This class provides methods 
    
    1. Search for catalogs in the given folder.
    
    2. Select catalogs with given criteria.
    
    3. Exclude catalogs with given criteria.
    
    4. Add catalogs to the set.
    
    5. Merge catalogs into single table.
        
    6. Select sources from each Catalog instance.
    
    """
    def __init__(self, catalogs: list[Catalog] = None):
        """
        Initialize the CatalogSet class.

        Parameters
        ----------
        catalogs : list[Catalog], optional
            List of Catalog instances. Defaults to None. If None, the CatalogSet will be dummy instance.
        """
        self.helper = Helper()
        self.catalogs = catalogs if catalogs is not None else []
        self.target_catalogs = self.catalogs
        self._last_filter = dict(
            file_key=None,
            filter=None,
            exptime=None,
            objname=None,
            obs_start=None,
            obs_end=None,
            seeing=None,
            depth=None,
            observatory=None,
            telname=None
        )
        self._last_mode = "select"  # <-- Track last mode (select or exclude)
    
    def __repr__(self):
        
        txt = f"CatalogSet[n_selected/n_catalogs= {len(self.target_catalogs)}/{len(self.catalogs)}] \n"
        txt += 'SELECT FILTER ============\n'
        for key, value in self._last_filter.items():
            prefix = "!" if self._last_mode == "exclude" and value is not None else ""
            txt += f"{prefix}{key:>11} = {value}\n"
        return txt
    
    def search_catalogs(self,
                        target_name: str,
                        search_key: str = '.cat',
                        folder: str = None,
                        recursive: bool = True,
                        n_proc: int = 16):
        """
        Search for catalogs in the given folder.
        
        Parameters
        ----------
        target_name : str
            Name of the target.
        search_key : str
            Search key for catalogs.
        folder : str, optional
            Folder to search for catalogs.
        recursive : bool, optional
            If True, search recursively in the given folder.
        n_proc : int, optional
            Number of processes to use for loading catalogs.

        Returns
        -------
        succeeded_catalogs : list[Catalog]
            List of Catalog instances that were successfully loaded.
        failed_catalogs : list[str]
            List of paths that failed to load.
        skipped_catalogs : list[str]
            List of paths that were skipped.
        """
        if folder is None:
            folder = self.helper.config['SCIDATA_DIR']
        folder = Path(folder)

        target_dir = next(folder.rglob(f"*{target_name}*"), None)
        print(f"Folder found: {target_dir}")
        if target_dir is None or not target_dir.is_dir():
            print(f"[WARNING] No matching subdirectory for target_name: {target_name}")
            return

        catalog_files = list(target_dir.rglob(f"*{search_key}")) if recursive else list(target_dir.glob(f"*{search_key}"))
        print(f"Catalog files: {len(catalog_files)}")

        existing_paths = [str(cat.path) for cat in self.catalogs]
        args = [(catalog_file, existing_paths) for catalog_file in catalog_files]

        succeeded_catalogs = []
        failed_catalogs = []
        skipped_catalogs = []

        with ProcessPoolExecutor(max_workers=n_proc) as executor:
            results = list(tqdm(executor.map(self._load_catalog_worker, args), total=len(args), desc="Loading catalogs"))

        for status, path, catalog in results:
            if status == 'success':
                succeeded_catalogs.append(catalog)
                print(f"[LOADED] {path}")
            elif status == 'skipped':
                skipped_catalogs.append(path)
                print(f"[SKIPPED] {path}")
            else:
                failed_catalogs.append(path)
                print(f"[FAILED] {path}")

        self.catalogs.extend(succeeded_catalogs)
        self.catalogs.sort(key=lambda x: x.path)

        return succeeded_catalogs, failed_catalogs, skipped_catalogs
    
    def merge_catalogs(self,
        max_distance_arcsec=1.0,
        ra_key='X_WORLD',
        dec_key='Y_WORLD',
        data_keys=['MAGSKY_AUTO', 'MAGERR_AUTO', 'MAGSKY_APER', 'MAGERR_APER',
                'MAGSKY_APER_1', 'MAGERR_APER_1', 'MAGSKY_APER_2', 'MAGERR_APER_2',
                'MAGSKY_APER_3', 'MAGERR_APER_3', 'MAGSKY_CIRC', 'MAGERR_CIRC'],
        join_type='outer'
        ):
        """
        Merge catalogs into single table.
        
        Parameters
        ----------
        max_distance_arcsec : float, optional
            Maximum distance in arcseconds for matching sources.
        ra_key : str, optional
            Column name for right ascension.
        dec_key : str, optional
            Column name for declination.
        data_keys : list[str], optional
            List of column names for data.
        join_type : str, optional
            Type of join to use.

        Returns
        -------
        merged_tbl : astropy.table.Table
            Merged table of catalogs.
        metadata : dict
            Metadata of catalogs.
        """
        from astropy.coordinates import SkyCoord
        import numpy as np
        import pandas as pd
        import astropy.units as u
        from tqdm import tqdm

        catalogs = self.target_catalogs
        dfs = []
        coords = []
        metadata = {}

        # Step 1: Load and preprocess all catalogs
        for i, catalog in tqdm(enumerate(catalogs), total=len(catalogs), desc="Preparing catalogs"):
            tbl = catalog.target_data.copy()
            ra = tbl[ra_key]
            dec = tbl[dec_key]
            mask = np.isfinite(ra) & np.isfinite(dec)
            tbl = tbl[mask]
            if len(tbl) == 0 or np.sum(np.isfinite(tbl[ra_key]) & np.isfinite(tbl[dec_key])) == 0:
                # Still create a dummy DataFrame with NaNs
                n_dummy = 1  # You can make this 1 or 0, depending on downstream needs
                row = {'ra': [0] * n_dummy, 'dec': [0] * n_dummy}
                for key in data_keys:
                    colname = f"{key}_idx{i}"
                    row[colname] = [np.nan] * n_dummy
                df = pd.DataFrame(row)
                df['catalog_id'] = i
                df['match_id'] = -1
                dfs.append(df)
                coords.append(SkyCoord([0]*n_dummy * u.deg, [0]*n_dummy * u.deg))  # dummy coords
                metadata[i] = catalog.info.to_dict()
                continue

            metadata[i] = catalog.info.to_dict()

            row = {'ra': tbl[ra_key], 'dec': tbl[dec_key]}
            for key in data_keys:
                colname = f"{key}_idx{i}"
                row[colname] = tbl[key] if key in tbl.colnames else np.full(len(tbl), np.nan)

            df = pd.DataFrame(row)
            df['catalog_id'] = i
            df['match_id'] = -1  # placeholder
            dfs.append(df)
            coords.append(SkyCoord(df['ra'].values * u.deg, df['dec'].values * u.deg))

        if len(dfs) == 0:
            return None, {}

        # Step 2: Initialize merged_df with first catalog
        merged_df = dfs[0].copy()
        merged_df['match_id'] = np.arange(len(merged_df))

        for i in tqdm(range(1, len(dfs)), desc="Merging catalogs"):
            c1 = SkyCoord(merged_df['ra'].values * u.deg, merged_df['dec'].values * u.deg)
            c2 = coords[i]
            df2 = dfs[i].copy()

            # Match c2 ? c1
            idx, d2d, _ = c2.match_to_catalog_sky(c1)
            sep_mask = d2d.arcsec < max_distance_arcsec
            df2.loc[sep_mask, 'match_id'] = merged_df.iloc[idx[sep_mask]]['match_id'].values

            matched = df2[df2['match_id'] >= 0].copy()
            unmatched = df2[df2['match_id'] < 0].copy()

            # Assign new match_id for unmatched
            if len(unmatched) > 0:
                unmatched['match_id'] = np.arange(
                    merged_df['match_id'].max() + 1,
                    merged_df['match_id'].max() + 1 + len(unmatched)
                )

            # Avoid duplicated column merge
            matched = matched[[col for col in matched.columns if col not in merged_df.columns or col == 'match_id']]
            merged_df = pd.merge(merged_df, matched, on='match_id', how=join_type)

            if join_type == 'outer' and len(unmatched) > 0:
                merged_df = pd.concat([merged_df, unmatched], ignore_index=True)

        # Step 3: Add detection count
        main_key = data_keys[0]
        match_cols = [col for col in merged_df.columns if col.startswith(main_key)]
        merged_df['n_detections'] = merged_df[match_cols].notna().sum(axis=1)
        # Remove columns that are completely NaN (dummy columns)
        idx_cols = [col for col in merged_df.columns if '_idx' in col]
        is_dummy_row = merged_df[idx_cols].isna().all(axis=1)
        merged_df = merged_df[~is_dummy_row].copy()
        merged_tbl = Table.from_pandas(merged_df)
        
        # Add coord column
        coord = SkyCoord(ra=merged_tbl['ra'] * u.deg, dec=merged_tbl['dec'] * u.deg)
        merged_tbl['coord'] = coord
        return merged_tbl, metadata

    def exclude_catalogs(self, 
                        file_key=None,
                        filter=None, 
                        exptime=None, 
                        objname=None, 
                        obs_start=None, 
                        obs_end=None,
                        seeing=None,
                        depth=None,
                        observatory=None,
                        telname=None):
        """
        Exclude catalogs that match the given criteria from self.catalogs.
        
        Select catalogs from self.catalogs and update self.target_catalogs.
        
        Parameters
        ----------
        file_key : str, optional
            File key to exclude.
        filter : str, optional
            Filter to exclude.
        exptime : float, optional
            Exposure time to exclude.
        objname : str, optional
            Object name to exclude.
        obs_start : str, optional
            Observation start time to exclude.
        obs_end : str, optional 
            Observation end time to exclude.
        seeing : float, optional
            Seeing to exclude.
        depth : float, optional
            Depth to exclude.
        observatory : str, optional
            Observatory to exclude.
        telname : str, optional
            Telescope name to exclude.

        Returns
        -------
        None
        """
        df = self.df

        # Convert inputs to arrays
        if file_key is not None:
            file_key = np.atleast_1d(file_key)
            for key in file_key:
                key = key.replace('*', '') if '*' in key else key
                df = df[~df['path'].str.contains(key)]

        if filter is not None:
            filter = np.atleast_1d(filter)
            df = df[~df['filter'].isin(filter)]
            
        if exptime is not None:
            exptime = np.atleast_1d(exptime)
            df = df[~df['exptime'].isin(exptime)]
            
        if objname is not None:
            objname = np.atleast_1d(objname)
            df = df[~df['objname'].isin(objname)]
            
        if obs_start is not None:
            obs_start = self.helper.flexible_time_parser(obs_start)
            df = df[Time(df['obsdate'].tolist()) < obs_start]
            
        if obs_end is not None:
            obs_end = self.helper.flexible_time_parser(obs_end)
            df = df[Time(df['obsdate'].tolist()) > obs_end]
            
        if seeing is not None:
            df = df[df['seeing'] >= seeing]
            
        if depth is not None:
            df = df[df['depth'] <= depth]
            
        if observatory is not None:
            observatory = np.atleast_1d(observatory)
            df = df[~df['observatory'].isin(observatory)]
            
        if telname is not None:
            telname = np.atleast_1d(telname)
            df = df[~df['telname'].isin(telname)]

        # Update target_catalogs
        if df.empty:
            self.target_catalogs = []
        else:
            self.target_catalogs = [self.catalogs[i] for i in df.index]

        self._last_filter = {
            'file_key': file_key,
            'filter': filter,
            'exptime': exptime,
            'objname': objname,
            'obs_start': obs_start,
            'obs_end': obs_end,
            'seeing': seeing,
            'depth': depth,
            'observatory': observatory,
            'telname': telname,
        }

        print(f"[INFO] Excluded catalogs based on given criteria. Remaining: {len(self.target_catalogs)}")

    def select_catalogs(self, 
                        file_key=None,
                        filter=None, 
                        exptime=None, 
                        objname=None, 
                        obs_start=None, 
                        obs_end=None,
                        seeing=None,
                        depth=None,
                        observatory=None,
                        telname=None):
        """
        Select catalogs that match the given criteria from self.catalogs.
        
        Select catalogs from self.catalogs and update self.target_catalogs.
        
        Parameters
        ----------
        file_key : str, optional
            File key to select.
        filter : str, optional
            Filter to select.   
        exptime : float, optional
            Exposure time to select.
        objname : str, optional
            Object name to select.
        obs_start : str, optional
            Observation start time to select.
        obs_end : str, optional
            Observation end time to select.
        seeing : float, optional
            Seeing to select.
        depth : float, optional
            Depth to select.
        observatory : str, optional
            Observatory to select.
        telname : str, optional
            Telescope name to select.

        Returns
        -------
        None
        """
        df = self.df

        # Convert inputs to arrays
        if file_key is not None:
            file_key = np.atleast_1d(file_key)
            for key in file_key:
                key = key.replace('*', '') if '*' in key else key
                df = df[df['path'].str.contains(key)]

        if filter is not None:
            filter = np.atleast_1d(filter)
            df = df[df['filter'].isin(filter)]
            
        if exptime is not None:
            exptime = np.atleast_1d(exptime)
            df = df[df['exptime'].isin(exptime)]
            
        if objname is not None:
            objname = np.atleast_1d(objname)
            df = df[df['objname'].isin(objname)]
            
        if obs_start is not None:
            obs_start = self.helper.flexible_time_parser(obs_start)
            df = df[Time(df['obsdate'].tolist()) >= obs_start]
            
        if obs_end is not None:
            obs_end = self.helper.flexible_time_parser(obs_end)
            df = df[Time(df['obsdate'].tolist()) <= obs_end]
            
        if seeing is not None:
            df = df[df['seeing'] < seeing]
            
        if depth is not None:
            df = df[df['depth'] > depth]
            
        if observatory is not None:
            observatory = np.atleast_1d(observatory)
            df = df[df['observatory'].isin(observatory)]
            
        if telname is not None:
            telname = np.atleast_1d(telname)
            df = df[df['telname'].isin(telname)]

        # Update target_catalogs
        if df.empty:
            self.target_catalogs = []
        else:
            self.target_catalogs = [self.catalogs[i] for i in df.index]

        self._last_filter = {
            'file_key': file_key,
            'filter': filter,
            'exptime': exptime,
            'objname': objname,
            'obs_start': obs_start,
            'obs_end': obs_end,
            'seeing': seeing,
            'depth': depth,
            'observatory': observatory,
            'telname': telname,
        }
        self._last_mode = "select"  # <-- mark as select

    def add_catalogs(self,
                     catalogs: Union[List[Catalog], Catalog]):
        """Add catalogs to the dataset."""
        if isinstance(catalogs, Catalog):
            catalogs = [catalogs]
            
        succeeded_catalogs = []
        failed_catalogs = []
        skipped_catalogs = []
        existing_paths = [cat.path for cat in self.catalogs]
        for catalog in catalogs:
            if catalog.path in existing_paths:
                print(f"[WARNING] Catalog already exists: {catalog.path}")
                skipped_catalogs.append(catalog.path)
                continue
            if not catalog.is_loaded:
                load_result = catalog.load_target_img(target_img = None)
                if load_result is False:
                    original_img = catalog.find_corresponding_fits()
                    if original_img is not None:
                        target_img = ScienceImage(original_img, telinfo = self.helper.estimate_telinfo(original_img), load=True)
                        load_result = catalog.load_target_img(target_img=target_img)
            if catalog.is_loaded:
                print(f"Loaded catalog: {catalog_file}")
                succeeded_catalogs.append(catalog)
            else:
                print(f"[ERROR] Failed to load catalog: {catalog_file}")
                failed_catalogs.append(catalog_file)
            
        self.catalogs.extend(succeeded_catalogs)
        self.catalogs.sort(key=lambda x: x.path)
        return succeeded_catalogs, failed_catalogs, skipped_catalogs

    def select_sources(self, ra, dec, radius = 60):
        """
        Select sources from all catalogs within the given radius around the input coordinates.
        
        Each catalog will be updated with the selected sources.
        
        Parameters
        ----------
        ra : float
            Right Ascension in degrees.
        dec : float
            Declination in degrees.
        radius : float
            Search radius in arcseconds.
        
        Returns
        -------
        None
        """
        results = []
        for cat in tqdm(self.target_catalogs, desc = 'Selecting sources...'):
            cat.select_sources(ra, dec, unit='coord', matching_radius=radius)
        
    @property
    def df(self):
        """
        Return a DataFrame containing metadata of all catalogs.
        """
        if len(self.catalogs) == 0:
            return pd.DataFrame()
        rows = []
        for cat in self.catalogs:
            info = cat.info
            rows.append({
                'catalog': cat,
                'path': info.path,
                'filter': info.filter,
                'exptime': info.exptime,
                'obsdate': info.obsdate,
                'observatory': info.observatory,
                'telname': info.telname,
                'objname': info.objname,
                'seeing': info.seeing,
                'depth': info.depth,
                'ra': info.ra,
                'dec': info.dec,
                'fov_ra': info.fov_ra,
                'fov_dec': info.fov_dec,
            })
        return pd.DataFrame(rows)

    def _load_catalog_worker(self, args):
        catalog_file, existing_paths = args
        try:
            if str(catalog_file) in existing_paths:
                return 'skipped', str(catalog_file), None

            catalog = Catalog(path=catalog_file, catalog_type='all', load=True)

            if not catalog.is_loaded:
                load_result = catalog.load_target_img(target_img=None)

            if catalog.is_loaded:
                return 'success', str(catalog_file), catalog
            else:
                return 'failed', str(catalog_file), None

        except Exception as e:
            return 'failed', str(catalog_file), None
        
    def _search_sources_worker(self, args):
        catalog, ra, dec, radius = args
        matched, sep, _ = catalog.search_sources(ra, dec, unit='coord', matching_radius=radius)
        if len(matched) > 0:
            return (catalog.info, matched, sep)
        return None

#%%
if __name__ == "__main__":
    self = CatalogSet()
    target_name = 'T01358'
    folder: str = None
    recursive: bool = True
    load: bool = True
    search_key = '100.fits.cat'  # Default search key for catalogs
    succeeded, failed, skipped = self.search_catalogs(
        target_name=target_name,
        folder=folder,
        recursive=recursive,
        search_key=search_key,
        n_proc = 16
    )
#%%
if __name__ == "__main__":
    self.select_sources(ra = 234.14732382, dec = 11.503295046, radius = 60)
#%%
if __name__ == "__main__":
    self.select_catalogs(
        filter = 'g'
    )
#%%
if __name__ == "__main__":
    result1= self.merge_catalogs_fast(
        ra_key='X_WORLD',
        dec_key='Y_WORLD',
        join_type = 'outer'
    )

# %%
if __name__ == '__main__':
    result = self.merge_catalogs(
        
    )
# %%



#%%
from pathlib import Path
from typing import List, Tuple, Optional
from astropy.io import ascii
from shapely.geometry import box
from concurrent.futures import ProcessPoolExecutor
import numpy as np

from ezphot.skycatalog import SkyCatalog


class SkyCatalogUtility:
    """
    Utility class for loading, filtering, and selecting sources from sky catalogs.
    
    Parameters
    ----------
    catalog_archive_path : str or Path, optional
        Path to the catalog archive summary file.
    default_mag_lower : float
        Default lower magnitude bound for filtering reference sources.
    default_mag_upper : float
        Default upper magnitude bound for filtering reference sources.
    """
    
    def __init__(self,
                 catalog_archive_path: Optional[str] = None):
        self.catalog_archive_path = Path(catalog_archive_path) if catalog_archive_path else None

    @staticmethod
    def _load_catalog_worker(row):
        """Worker to load a SkyCatalog from a summary row."""
        return SkyCatalog(objname=row['objname'],
                          catalog_type=row['cat_type'],
                          fov_ra=row['fov_ra'],
                          fov_dec=row['fov_dec'])

    def select_reference_sources(self,
                                  catalog: SkyCatalog,
                                  mag_lower: Optional[float] = None,
                                  mag_upper: Optional[float] = None,
                                  **kwargs) -> Tuple['astropy.table.Table', list]:
        """
        Select reference sources from a given SkyCatalog based on predefined cuts.
        """
        if not catalog.data:
            raise RuntimeError(f'No catalog data found for {catalog.objname}')

        # Define default cutlines
        cutlines = {
            'APASS': dict(e_V_mag=[0.001, 0.2], V_mag=[mag_lower, mag_upper]),
            'GAIA': dict(V_flag=[0, 1], V_mag=[mag_lower, mag_upper]),
            'GAIAXP': {"pmra": [-20, 20], "pmdec": [-20, 20], "bp-rp": [0.0, 1.5], "g_mean": [mag_lower, mag_upper]},
            'PS1': {"gFlags": [0, 10], "g_mag": [mag_lower, mag_upper]},
            'SMSS': {"ngood": [5, 999], "class_star": [0.3, 1.0], "g_mag": [mag_lower, mag_upper]}
        }

        if catalog.catalog_type not in cutlines:
            raise ValueError(f"Invalid catalog type: {catalog.catalog_type}")

        cutline = {**cutlines[catalog.catalog_type], **kwargs}

        ref_sources = catalog.data
        applied_kwargs = []
        for key, value in cutline.items():
            if key in ref_sources.colnames:
                applied_kwargs.append({key: [value]})
                ref_sources = ref_sources[(ref_sources[key] >= value[0]) & (ref_sources[key] <= value[1])]

        return ref_sources, applied_kwargs

    def get_catalogs_coord(self,
                           ra: float,
                           dec: float,
                           objname: str = None,
                           fov_ra: float = 1.0,
                           fov_dec: float = 1.0,
                           catalog_type: str = 'GAIAXP',
                           fraction_criteria: float = 0.05,
                           query_when_not_archived: bool = True,
                           verbose: bool = False) -> List[SkyCatalog]:
        """
        Load catalogs that overlap with a given coordinate and FOV.
        """
        def make_sky_rectangle(ra_center, dec_center, fov_ra, fov_dec):
            ra_offset = (fov_ra / 2) / np.cos(np.radians(dec_center))
            dec_offset = fov_dec / 2
            return box(ra_center - ra_offset, dec_center - dec_offset,
                       ra_center + ra_offset, dec_center + dec_offset)

        # Load catalog summary
        catalog_path = self.catalog_archive_path
        if catalog_path is None:
            catalog_path = Path(__file__).parent / 'catalog_archive' / 'catalog_summary.ascii_fixed_width'
        summary = ascii.read(catalog_path, format='fixed_width')
        summary = summary[summary['cat_type'] == catalog_type]

        # Fast angular pre-filter
        ra_arr = np.asarray(summary['ra'])
        dec_arr = np.asarray(summary['dec'])
        cos_dec0, sin_dec0 = np.cos(np.radians(dec)), np.sin(np.radians(dec))
        cos_dec, sin_dec = np.cos(np.radians(dec_arr)), np.sin(np.radians(dec_arr))
        delta_ra = np.radians(ra_arr - ra)
        angular_sep = np.degrees(np.arccos(np.clip(sin_dec0 * sin_dec + cos_dec0 * cos_dec * np.cos(delta_ra), -1, 1)))

        search_radius = min(5, 10 * np.sqrt(fov_ra ** 2 + fov_dec ** 2))
        summary_nearby = summary[angular_sep < search_radius]

        # Target polygon
        target_poly = make_sky_rectangle(ra, dec, fov_ra, fov_dec)

        # Intersection filter
        filtered_rows = []
        for row in summary_nearby:
            tile_poly = make_sky_rectangle(row['ra'], row['dec'], row['fov_ra'], row['fov_dec'])
            if tile_poly.intersects(target_poly):
                intersection_fraction = tile_poly.intersection(target_poly).area / target_poly.area
                print(f"Catalog {row['objname']} intersection fraction: {intersection_fraction:.2f}")
                if intersection_fraction >= fraction_criteria:
                    filtered_rows.append(row)

        # Parallel load
        with ProcessPoolExecutor(max_workers=6) as executor:
            catalogs = list(executor.map(self._load_catalog_worker, filtered_rows))

        if verbose:
            if catalogs:
                print(f"Found {len(catalogs)} catalogs matching the region (fraction > {fraction_criteria}).")
            else:
                print("No catalogs found.")
        
        if query_when_not_archived:
            if len(catalogs) == 0:
                skycatalog = SkyCatalog(objname = objname, ra = ra, dec = dec, fov_ra = fov_ra, fov_dec = fov_dec, catalog_type = catalog_type, overlapped_fraction = fraction_criteria)
                catalogs = [skycatalog]

        return catalogs
# %%
if __name__ == "__main__":
    # Example usage
    from ezphot.utils import DataBrowser
    from ezphot.imageobjects import *
    from ezphot.methods import *
    import glob
    databrowser = DataBrowser('scidata')
    databrowser.observatory = 'RASA36'
    databrowser.telkey = 'RASA36_KL4040_HIGH_1x1'
    databrowser.objname = 'NGC1566'
    stacked_imgset = databrowser.search(pattern='Calib*60.com.fits', return_type='science')
    stacked_imglist = stacked_imgset.target_images
#%%
    # db = DataBrowser('scidata')
    # db.objname = 'NGC1566'
    # db.observatory = 'RASA36'
    
    # target_imgset = db.search('*.com.fits', 'science')
    # target_imglist = target_imgset.target_images
    target_img = stacked_imglist[0]
    ra = target_img.ra
    dec = target_img.dec
    fov_ra = target_img.fovx
    fov_dec = target_img.fovy
    catalog_type = 'APASS'
    utility = SkyCatalogUtility()
    
    catalogs = utility.get_catalogs_coord(ra, dec, fov_ra, fov_dec, catalog_type=catalog_type, verbose=True)

# %%




# %%
from astropy.io import fits
from astropy.io import ascii
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy.table import Table

from tqdm import tqdm
import os
import numpy as np
from astropy.table import unique
import inspect
from astropy import constants as const

# %%

class SpectroscopyHelper():

    def __init__(self):
        self.c = const.c.value
        self.sigma = const.sigma_sb.cgs.value
        self.d10pc = 10 * const.pc.cgs.value
        self.c = const.c.cgs.value
        self.h = const.h.cgs.value
        self.k = const.k_B.cgs.value
        
    @property
    def specpath(self):
        # Get the file where the class is defined
        file_path = inspect.getfile(SpectroscopyHelper)

        # Convert the file path to an absolute path using os.path.abspath
        absolute_path = os.path.abspath(file_path)

        path_dir = os.path.join(os.path.dirname(absolute_path),'../spectroscopy')

        return path_dir

    def __repr__(self):
        methods = [f'SpectroscopyHelper.{name}()\n' for name, method in inspect.getmembers(
            SpectroscopyHelper, predicate=inspect.isfunction) if not name.startswith('_')]
        txt = '[Methods]\n'+''.join(methods)
        return txt

# %%



#%%
import logging
from pathlib import Path
from typing import Union
import numpy as np
from astropy.io import fits
from astropy.time import Time
from astropy.wcs import WCS
from astropy.wcs.utils import proj_plane_pixel_scales
from astropy.io.fits import Header

from ezphot.configuration import TIPConfig
from ezphot.helper import Helper
#%%

class LazyFileHandler(logging.FileHandler):
    def __init__(self, filename, mode='a', encoding=None, delay=True):
        # `delay=True` avoids creating the file immediately
        super().__init__(filename, mode, encoding, delay=delay)
    
class Logger:
    def __init__(self, logger_name):
        self.path = logger_name
        self._log = self.createlogger(logger_name)

    def log(self):
        return self._log

    def createlogger(self, logger_name, logger_level='INFO'):
        logger = logging.getLogger(logger_name)
        if len(logger.handlers) > 0:
            return logger  # Logger already exists

        logger.setLevel(logger_level)
        formatter = logging.Formatter(
            datefmt='%Y-%m-%d %H:%M:%S',
            fmt='[%(levelname)s] %(asctime)-15s | %(message)s'
        )

        # Stream Handler
        streamHandler = logging.StreamHandler()
        streamHandler.setLevel(logger_level)
        streamHandler.setFormatter(formatter)
        logger.addHandler(streamHandler)

        # Lazy File Handler
        fileHandler = LazyFileHandler(filename=logger_name, delay=True)
        fileHandler.setLevel(logger_level)
        fileHandler.setFormatter(formatter)
        logger.addHandler(fileHandler)

        return logger
    
#%%
class BaseImage(TIPConfig):
    """Base class for loading, managing, and inspecting FITS images.

    Provides lazy-loading of FITS data and header, WCS access, metadata extraction,
    and visualization support. Used as the parent class for all telescope-acquired images
    such as science, reference, and calibration frames.

    Parameters
    ----------
    path : Union[str, Path]
        Path to the FITS image file.
    telinfo : dict, optional
        Telescope information dictionary. If not provided, it will be estimated automatically.
    """

    def __init__(self, path: Union[Path, str], telinfo : dict = None):
        path = Path(path)

        self.helper = Helper()
        self.path = path
        if telinfo is None:
            telinfo = self.helper.estimate_telinfo(self.path)
        self.telinfo = telinfo
        self.telkey = self._get_telkey()
        # Initialize or load status
        super().__init__(telkey = self.telkey)

        self._data = None
        self._header = Header()

    def rename(self, new_name: str):
        """Rename the image file and update the internal path.
        
        Parameters
        ----------
        new_name : str
            New name of the image file.
        
        Returns
        -------
        None
        """
        old_path = self.path
        new_path = self.path.parent / new_name

        # If the target exists, remove it (overwrite)
        if new_path.exists():
            new_path.unlink()  # remove the existing file

        old_path.rename(new_path)
        self.path = new_path
        print(f"Renamed {old_path} to {new_path}")

    def clear(self, clear_data: bool = True, clear_header: bool = False):
        """Clear the image data and/or header from memory.
        
        Parameters
        ----------
        clear_data : bool, optional
            If True, clear the image data. Default is True.
        clear_header : bool, optional
            If True, clear the image header. Default is False.
        """
        if clear_data:
            self._data = None
        if clear_header:
            self._header = Header()

    def update_header(self, **kwargs):
        """Update FITS header values using known key variants.
        
        Parameters
        ----------
        kwargs : dict
            Keyword arguments to update the FITS header.
        
        Returns
        -------
        None
        """
        if self._header is None:
            print("WARNING: Header is not loaded. Cannot update.")
            return
        else:
            for key, value in kwargs.items():
                key_upper = key.upper()
                if key_upper in self._key_variants.keys():
                    key_variants = self._key_variants[key_upper]
                    for key_variant in key_variants:
                        if key_variant in self._header:
                            self._header[key_variant] = value
                else:
                    print(f'WARNING: Key {key} not found in key_variants.')

    def load(self):
        """Load both image data and header from disk.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        if not self.is_exists:
            raise FileNotFoundError(f'File not found: {self.path}')
        self.data
        self.header
        
    def show(self, 
             cmap='gray', 
             scale='zscale', 
             downsample=4, 
             figsize=(8, 6), 
             title=None, 
             save_path: str = False,
             close_fig: bool = False):
        """
        Visualize the FITS image using slicing-based downsampling and scaling.

        Parameters
        ----------
        cmap : str, optional
            Matplotlib colormap. Default is 'gray'.
        scale : str, optional
            Scaling method. Default is 'zscale'.
        downsample : int, optional
            Step size for downsampling via slicing. Default is 4.
        figsize : tuple, optional
            Matplotlib figure size. Default is (8, 6).
        title : str, optional
            Plot title. Default is None.
        save_path : str, optional
            Path to save the figure. Default is None.
        close_fig : bool, optional
            If True, close the figure after saving. Default is False.
        """
        import matplotlib.pyplot as plt
        from astropy.visualization import ZScaleInterval, MinMaxInterval
        from mpl_toolkits.axes_grid1 import make_axes_locatable

        data = self.data
        if data is None:
            print("WARNING: Image data is not loaded. Please load the image first.")
            return
        
        # Handle NaN and inf
        data = np.nan_to_num(data, nan=0.0, posinf=0.0, neginf=0.0)

        # Store original shape before slicing
        ny, nx = data.shape

        # Downsampling using slicing
        if downsample > 1:
            data = data[::downsample, ::downsample]

        # Scaling
        if scale == 'zscale':
            interval = ZScaleInterval()
        elif scale == 'minmax':
            interval = MinMaxInterval()
        else:
            print(f"Invalid scale option: {scale}. Use 'zscale' or 'minmax'.")
            return

        vmin, vmax = interval.get_limits(data)

        # Plot image
        fig, ax = plt.subplots(figsize=figsize)
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.05)  # colorbar axis

        img = ax.imshow(data, cmap=cmap, origin='lower', vmin=vmin, vmax=vmax)

        # Set original pixel ticks
        yticks = np.linspace(0, data.shape[0]-1, num=6, dtype=int)
        xticks = np.linspace(0, data.shape[1]-1, num=6, dtype=int)
        ax.set_yticks(yticks)
        ax.set_yticklabels((yticks * downsample).astype(int))
        ax.set_xticks(xticks)
        ax.set_xticklabels((xticks * downsample).astype(int))

        ax.set_title(title or self.path.name)
        fig.colorbar(img, cax=cax, label='Pixel value')  # colorbar on the matched axis

        plt.tight_layout()
        if save_path:
            plt.savefig(save_path, dpi=300)
            print(f"Saved: {save_path}")
        
        if close_fig:
            plt.close(fig)
        
        return fig, ax
        
    def show_position(self,
                    x: float,
                    y: float,
                    coord_type: str = 'pixel',
                    downsample: int = 4,
                    zoom_radius_pixel: int = 100,
                    cmap: str = 'gray',
                    scale: str = 'zscale',
                    figsize=(6, 6),
                    ax=None,
                    save_path: str = None,
                    title: bool = True):
        """
        Display a zoomed-in view around a given (x, y) or (RA, Dec) position.

        Parameters
        ----------
        x : float
            X-coordinate (pixel or RA)
        y : float
            Y-coordinate (pixel or Dec)
        coord_type : str, optional
            'pixel' or 'coord'. Default is 'pixel'.
        downsample : int, optional
            Step size for downsampling via slicing. Default is 4.
        zoom_radius_pixel : int, optional
            Radius of the zoom-in region in pixels. Default is 100.
        cmap : str, optional
            Matplotlib colormap. Default is 'gray'.
        scale : str, optional
            Scaling method. Default is 'zscale'.
        figsize : tuple, optional
            Matplotlib figure size. Default is (6, 6).
        ax : matplotlib Axes object, optional
            If provided, the image will be plotted on this axis. Default is None.
        save_path : str, optional
            Path to save the figure. Default is None.
        title : bool, optional
            Whether to display the title. Default is True.
        """
        import matplotlib.pyplot as plt
        from astropy.visualization import ZScaleInterval, MinMaxInterval
        from matplotlib.patches import Circle

        data = self.data
        wcs = self.wcs
        if data is None:
            print("No image data loaded.")
            return

        # Convert (RA, Dec) to pixel if needed
        if coord_type == 'coord':
            if wcs is None:
                print("No valid WCS for sky-to-pixel conversion.")
                return
            from astropy.coordinates import SkyCoord
            import astropy.units as u
            coord = SkyCoord(x * u.deg, y * u.deg, frame='icrs')
            x_pix, y_pix = wcs.world_to_pixel(coord)
        elif coord_type == 'pixel':
            x_pix, y_pix = x, y
        else:
            raise ValueError("coord_type must be 'pixel' or 'coord'.")

        x_pix, y_pix = int(x_pix), int(y_pix)

        # Extract zoom window
        x_min = max(0, x_pix - zoom_radius_pixel)
        x_max = min(data.shape[1], x_pix + zoom_radius_pixel)
        y_min = max(0, y_pix - zoom_radius_pixel)
        y_max = min(data.shape[0], y_pix + zoom_radius_pixel)
        size = x_max - x_min
        cutout = data[y_min:y_max:downsample, x_min:x_max:downsample]

        # Scaling
        interval = ZScaleInterval() if scale == 'zscale' else MinMaxInterval()
        vmin, vmax = interval.get_limits(cutout)

        # Draw
        if ax is None:
            fig, ax = plt.subplots(figsize=figsize)
        else:
            fig = ax.figure

        ax.imshow(cutout, cmap=cmap, origin='lower', vmin=vmin, vmax=vmax)
        ax.add_patch(Circle(
            ((x_pix - x_min) // downsample, (y_pix - y_min) // downsample),
            radius=size * (0.08/downsample), edgecolor='red', facecolor='none', linewidth=0.5
        ))
        ax.axis('off')
        ax.set_aspect('auto')  # ? Avoid square enforcement

        if title is not None:
            ax.set_title(f"{title}", fontsize=8, pad=1)  # Less padding

        if save_path:
            fig.savefig(save_path, dpi=300)
            print(f"Saved: {save_path}")

        return fig, ax

    def run_ds9(self):
        """Open the image with SAOImage DS9.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self.helper.run_ds9(self.path)        
        
    @property
    def data(self):
        """Image data array."""
        if not self.is_data_loaded and self.is_exists:
            try:
                self._data = fits.getdata(self.path)
            except Exception as e:
                print(f"Failed to load data from {self.path}: {e}")
        return self._data

    @data.setter
    def data(self, value):
        self._data = value

    @property
    def header(self):
        """FITS header object."""
        if not self.is_header_loaded and self.is_exists:
            try:
                self._header = fits.getheader(self.path)
            except Exception as e:
                print(f"Failed to load header from {self.path}: {e}")
        return self._header

    @header.setter
    def header(self, value):
        self._header = value
         
    @property
    def is_data_loaded(self):
        """Whether the image data is loaded."""
        return self._data is not None

    @property
    def is_header_loaded(self):
        """Whether the image header is loaded."""
        return isinstance(self._header, Header) and len(self._header) > 0

    @property
    def is_exists(self):
        """Whether the image file exists."""
        return self.path.exists()
        
    @property
    def observatory(self):
        """ Name of the observatory."""
        return str(self.telinfo['obs'])
    
    @property
    def telname(self):
        """ Name of the telescope."""
        header = self._header
        for key in self._key_variants['TELESCOP']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def imgtype(self):
        """ Type of the image. Among BIAS, DARK, FLAT, LIGHT, OBJECT, UNKNOWN."""
        header = self._header
        for key in self._key_variants['IMGTYPE']:
            if key in header:
                imgtype = header[key]
                imgtype_variants = dict(BIAS= ['BIAS', 'Bias', 'bias', 'ZERO', 'Zero', 'zero'],
                                        DARK= ['DARK', 'Dark', 'dark'],
                                        FLAT= ['FLAT', 'Flat', 'flat'],
                                        LIGHT= ['LIGHT', 'Light', 'light', 'OBJECT', 'Object', 'object'])
                for key, variants in imgtype_variants.items():
                    if imgtype in variants:
                        return key
        print('WARNING: IMGTYPE not found in header')
        return 'UNKNOWN'
    
    @property
    def altitude(self):
        """ Altitude of the telescope.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        altitude : float
            Altitude of the telescope.
        """
        header = self._header
        for key in self._key_variants['ALTITUDE']:
            if key in header:
                return float(header[key])
        return None

    
    @property
    def azimuth(self):
        """Azimuth of the telescope."""
        header = self._header
        for key in self._key_variants['AZIMUTH']:
            if key in header:
                return float(header[key])
        return None

    @property 
    def ra(self):
        """Right ascension of the target."""
        header = self._header
        for key in self._key_variants['RA']:
            if key in header:
                return float(header[key])
        return None
        
    @property
    def dec(self):
        """Declination of the target."""
        header = self._header
        for key in self._key_variants['DEC']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def objname(self):
        """Object name."""
        header = self._header
        for key in self._key_variants['OBJECT']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def obsmode(self):
        """Observation mode."""
        header = self._header
        for key in self._key_variants['OBSMODE']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def specmode(self):
        """Spectroscopic mode."""
        header = self._header
        for key in self._key_variants['SPECMODE']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def filter(self):
        """Filter name."""
        header = self._header
        for key in self._key_variants['FILTER']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def ccdtemp(self):
        """CCD temperature."""
        header = self._header
        for key in self._key_variants['CCD-TEMP']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def obsdate(self):
        """Observation date in UTC."""
        header = self._header
        # First, search for utcdate
        for key in self._key_variants['DATE-OBS']:
            if key in header:
                return Time(header[key], format = 'isot', scale = 'utc').isot
        # If not found, search for jd
        for key in self._key_variants['JD']:
            if key in header:
                return Time(header[key], format = 'jd').isot
        # If not found, search for mjd
        for key in self._key_variants['MJD']:
            if key in header:
                return Time(header[key], format = 'mjd').isot
        return None
    
    @property
    def mjd(self):
        """Modified Julian date of the observation."""
        header = self._header
        for key in self._key_variants['MJD']:
            if key in header:
                return float(header[key])
        for key in self._key_variants['JD']:
            if key in header:
                return Time(header[key], format = 'jd').mjd
        for key in self._key_variants['DATE-OBS']:
            if key in header:
                return Time(header[key], format = 'isot', scale = 'utc').mjd
        return None
    
    @property
    def jd(self):
        """Julian date of the observation."""
        header = self._header
        for key in self._key_variants['JD']:
            if key in header:
                return float(header[key])
        for key in self._key_variants['MJD']:
            if key in header:
                return Time(header[key], format = 'mjd').jd
        for key in self._key_variants['DATE-OBS']:
            if key in header:
                return Time(header[key], format = 'isot', scale = 'utc').jd
        return None

    @property
    def exptime(self):
        """Exposure time of the image."""
        header = self._header
        for key in self._key_variants['EXPTIME']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def binning(self):
        """Binning of the image."""
        header = self._header
        for key in self._key_variants['BINNING']:
            if key in header:
                return int(header[key])
        return None
        
    @property
    def gain(self):
        """Gain of the image."""
        header = self._header
        for key in self._key_variants['GAIN']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def egain(self):
        """Electron gain of the image."""
        header = self._header
        for key in self._key_variants['EGAIN']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def naxis1(self):
        """Number of pixels along the first axis."""
        header = self._header
        for key in self._key_variants['NAXIS1']:
            if key in header:
                return int(header[key])
        return None

    @property
    def naxis2(self):
        """Number of pixels along the second axis."""
        header = self._header
        for key in self._key_variants['NAXIS2']:
            if key in header:
                return int(header[key])
        return None
    
    @property
    def ncombine(self):
        """Number of combined images."""
        header = self._header
        for key in self._key_variants['NCOMBINE']:
            if key in header:
                return int(header[key])
        return None
    
    @property
    def biaspath(self):
        """Path to the bias image."""
        header = self._header
        for key in self._key_variants['BIASPATH']:
            if key in header:
                return str(header[key])
        return None
            
    @property
    def darkpath(self):
        """Path to the dark image."""
        header = self._header
        for key in self._key_variants['DARKPATH']:
            if key in header:
                return str(header[key])
        return None

    @property
    def flatpath(self):
        """Path to the flat image."""
        header = self._header
        for key in self._key_variants['FLATPATH']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def maskpath(self):
        """Path to the mask image."""
        header = self._header
        for key in self._key_variants['MASKPATH']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def masktype(self):
        """Type of the mask image."""
        header = self._header
        for key in self._key_variants['MASKTYPE']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def bkgpath(self):
        """Path to the background image."""
        header = self._header
        for key in self._key_variants['BKGPATH']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def bkgtype(self):
        """Type of the background image."""
        header = self._header
        for key in self._key_variants['BKGTYPE']:
            if key in header:
                return str(header[key])
        return None
    
    @property
    def emappath(self):
        """Path to the emap image."""
        header = self._header
        for key in self._key_variants['EMAPPATH']:
            if key in header:
                return str(header[key])
        return None

    @property
    def emaptype(self):
        """Type of the emap image."""
        header = self._header
        for key in self._key_variants['EMAPTYPE']:
            if key in header:
                return str(header[key])
        return None
            
    @property
    def fovx(self):
        """Field of view along the first axis."""
        pixscale = self.pixelscale
        if pixscale is not None:
            fovx = pixscale[0] * self.naxis1 / 3600  # Convert to degrees
            return float('%.3f' % fovx)
        else:
            return float('%.3f' %(self.telinfo['pixelscale'] * self.telinfo['x'] / 3600))
    
    @property
    def fovy(self):
        """Field of view along the second axis."""
        pixscale = self.pixelscale
        if pixscale is not None:
            fovy = pixscale[1] * self.naxis2 / 3600  # Convert to degrees
            return float('%.3f' % fovy)
        else:
            return float('%.3f' %(self.telinfo['pixelscale'] * self.telinfo['y'] / 3600))
    
    @property
    def wcs(self):
        """WCS information of the image."""
        try:
            wcs = WCS(self._header)
            return wcs
        except:
            return None
    
    @property
    def center(self):
        """Center pixel (0-based) and its world coordinates (RA, Dec)."""
        x_center = (self.naxis1 - 1) / 2
        y_center = (self.naxis2 - 1) / 2
        ra = dec = None

        if self.wcs is not None:
            try:
                skycoord = self.wcs.pixel_to_world(x_center, y_center)
                ra = skycoord.ra.deg
                dec = skycoord.dec.deg
            except Exception as e:
                print(f"WCS conversion failed: {e}")

        return {'x': x_center, 'y': y_center, 'ra': ra, 'dec': dec}
        
    @property
    def pixelscale(self):
        """Pixel scale of the image."""
        try:
            return proj_plane_pixel_scales(self.wcs) * 3600  # Convert to arcseconds
        except:
            return None
    
    @property
    def zp(self):
        """Zero point of the image."""
        header = self._header
        for key in self._key_variants['ZP']:
            if key in header:
                return float(header[key])
        return None

    @property
    def zperr(self):    
        """Zero point error of the image."""
        header = self._header
        for key in self._key_variants['ZPERR']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def depth(self):
        """Depth of the image."""
        header = self._header
        for key in self._key_variants['DEPTH']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def seeing(self):
        """Seeing of the image."""
        header = self._header
        for key in self._key_variants['SEEING']:
            if key in header:
                return float(header[key])
        return None
    
    @property
    def ncombine(self):
        """Number of combined images."""
        header = self._header
        for key in self._key_variants['NCOMBINE']:
            if key in header:
                return int(header[key])
        return None
        
    @property
    def _key_variants(self):
        '''Key variants of the FITS header.'''
        # Define key variants, if a word is duplicated in the same variant, posit the word with the highest priority first
        key_variants_upper = {
            # Observation information
            'ALTITUDE': ['ALT', 'ALTITUDE', 'CENTALT'],
            'AZIMUTH': ['AZ', 'AZIMUTH', 'CENTAZ'],
            'GAIN': ['GAIN'],
            'EGAIN': ['EGAIN'],
            'CCD-TEMP': ['CCDTEMP', 'CCD-TEMP'],
            'FILTER': ['FILTER', 'FILTNAME', 'BAND'],
            'IMGTYPE': ['IMGTYPE', 'IMAGETYP', 'IMGTYP'],
            'EXPTIME': ['EXPTIME', 'EXPOSURE'],
            'DATE-OBS': ['DATE-OBS', 'OBSDATE', 'UTCDATE'],
            'DATE-LOC': ['DATE-LOC', 'DATE-LTC', 'LOCDATE', 'LTCDATE'],
            'JD' : ['JD', 'JD-HELIO', 'JD-UTC', 'JD-OBS'],
            'MJD' : ['MJD', 'MJD-HELIO', 'MJD-UTC', 'MJD-OBS'],
            'RA': ['CRVAL1', 'RA', 'OBJCTRA', 'OBSRA'],
            'DEC': ['CRVAL2', 'DEC', 'DECL', 'DEC.', 'DECL.', 'CRVAL2', 'OBJCTDEC', 'OBSDEC'],   
            'TELESCOP' : ['TELESCOP', 'TELNAME'],
            'BINNING': ['BINNING', 'XBINNING'],
            'OBJECT': ['OBJECT', 'OBJNAME', 'TARGET', 'TARNAME'],
            'OBJCTID': ['OBJCTID', 'OBJID', 'ID'],
            'OBSMODE': ['OBSMODE', 'MODE'],
            'SPECMODE': ['SPECMODE'],
            'NTELESCOP': ['NTELESCOP', 'NTEL'],
            'NCOMBINE': ['NCOMBINE', 'NCOMB'],
            'NOTE': ['NOTE'],
            'NAXIS1': ['NAXIS1'],
            'NAXIS2': ['NAXIS2'],
            # Additional key after processing
            'CTYPE1': ['CTYPE1'],
            'CTYPE2': ['CTYPE2'],
            'CRVAL1': ['CRVAL1'],
            'CRVAL2': ['CRVAL2'],
            'SEEING': ['SEEING'],
            'ELONGATION': ['ELONGATION', 'ELONG'],
            'SKYSIG': ['SKYSIG', 'SKY_SIG'],
            'SKYVAL': ['SKYVAL', 'SKY_VAL'],
            'ZP': ['ZP_AUTO', 'ZP_2'],
            'ZPERR': ['ZPERR_AUTO', 'EZP_2'],
            'DEPTH': ['UL5_APER_2', 'UL5_2'],
            # Path information
            'SAVEPATH': ['SAVEPATH'],
            'BIASPATH': ['BIASPATH'],
            'DARKPATH': ['DARKPATH'],
            'FLATPATH': ['FLATPATH'],
            'BKGPATH': ['BKGPATH'],
            'BKGTYPE': ['BKGTYPE'],
            'EMAPPATH': ['EMAPPATH'],
            'EMAPTYPE': ['EMAPTYPE'],
            'MASKPATH': ['MASKPATH'],
            'MASKTYPE': ['MASKTYPE'],
        }

        return key_variants_upper

    def _get_telkey(self):
        """ Get the telescope name from the FITS header """
        telinfo = self.telinfo
        if telinfo['mode']:
            telkey = f"{telinfo['obs']}_{telinfo['ccd']}_{telinfo['mode']}_{telinfo['binning']}x{telinfo['binning']}"
        else:
            telkey = f"{telinfo['obs']}_{telinfo['ccd']}_{telinfo['binning']}x{telinfo['binning']}"
        return telkey

# %%